/*
 * spinc - Convert SPIN objects to C arrays. Based on the original program 
 *         by Steve Densen (original copyright below). 
 *
 *         Modified by Ross Higson for use with Catalina.
 *
 *         See spinc.h for version history.
 *
 * ----------------------------------------------------------------------------
 * @file spinc.h
 * Defines data structures and API for spinc converter
 * Copyright (c) 2008, Steve Denson
 * ----------------------------------------------------------------------------
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "spinc.h"

#define DEFAULT_STACK_SIZE 80 // 80 bytes (20 longs)

static int    named = 0;
static int    verbose = 0;
static int    diagnose = 0;
static int    tiny_only = 0;
static int    long_array = 1;
static int    func_named = 0;
static int    all_objects = 0;
static int    gen_function = 0;
static char   func_name[MAX_LINELEN+1];
static char   array_name[MAX_LINELEN+1];
static int    stack_size = DEFAULT_STACK_SIZE;

static int    file_count = 0;                  
static char * file_name[MAX_FILES];


// safecpy will never write more than size characters, 
// and is guaranteed to null terminate its result, so
// make sure the buffer passed is at least size + 1
char * safecpy(char *dst, const char *src, size_t size) {
   dst[size] = '\0';
   if (src) {
      return strncpy(dst, src, size - strlen(dst));
   }
   return NULL;
}


// safecat will never write more than size characters, 
// and is guaranteed to null terminate its resul, so
// make sure the buffer passed is at least size + 1
char * safecat(char *dst, const char *src, size_t size) {
   dst[size] = '\0';
   if (src) {
      return strncat(dst, src, size - strlen(dst));
   }
   return NULL;
}


// pathcat will check for paths with spaces and quote
// them if we are using Win32 style path names, otherwise
// it is much the same as two safecats
void pathcat(char *dst, const char *src, const char *sfx, size_t max) {
//#ifdef WIN32_PATHS
   if ((strchr(src, ' ') != NULL) && (src[0] != '\"')) {
      safecat(dst, "\"", max);
      safecat(dst, src, max);
      if (sfx != NULL) {
         safecat(dst, sfx, max);
      }
      safecat(dst, "\"", max);
   }
   else {
      safecat(dst, src, max);
      if (sfx != NULL) {
         safecat(dst, sfx, max);
      }
   }
//#else
//   safecat(dst, src, max);
//   if (sfx != NULL) {
//      safecat(dst, sfx, max);
//   }
//#endif   
}


void help(char *my_name) {
   int i;

   fprintf(stderr, "Convert Compiled Spin Programs to C arrays.\n");
   fprintf(stderr, "\nusage: %s [options] binary_file(s)\n\n", my_name);
   fprintf(stderr, "options:  -? or -h  print this helpful message and exit\n");
   fprintf(stderr, "          -a        include all SPIN objects (default is first only)\n");
   fprintf(stderr, "          -b        generate array of bytes (default is array of longs)\n");
   fprintf(stderr, "          -c        generate callable C function (TINY, SMALL or LARGE)\n");
   fprintf(stderr, "          -d        diagnostic mode\n");
   fprintf(stderr, "          -f func   use 'func' for the spin function (default '_coginit_Spin')\n");
   fprintf(stderr, "          -l        generate array of longs (default)\n");
   fprintf(stderr, "          -n name   use 'name' for arrays (default is binary file name)\n");
   fprintf(stderr, "          -s size   size (in bytes) of stack (default is %d bytes)\n", DEFAULT_STACK_SIZE);
   fprintf(stderr, "          -t        generate callable C function (TINY mode only)\n");
   fprintf(stderr, "          -v        verbose mode\n");
}

/*
 * decode arguments, building file list - return -1 if
 * there is no further processing to do
 */
int decode_arguments (int argc, char *argv[]) {
   int  i = 0;
   int  code = 0;
   char filename[MAX_LINELEN + 3 + 1];

   fprintf(stderr, "Catalina SpinC %s\n", VERSION); 

   while ((code >= 0) && (argc--)) {
      if (diagnose) {
         fprintf(stderr, "arg: %s\n", argv[i]);
      }
      if (i > 0) {
         if (argv[i][0] == '-') {
            if (diagnose) {
               fprintf(stderr, "switch: %s\n", argv[i]);
            }
            // it's a command line switch
            switch (argv[i][1]) {
               case 'h':
               case '?':
                  if (strlen(argv[0]) == 0) {
                     // in case my name was not passed in ...
                     help("spinc");
                  }
                  else {
                     help(argv[0]);
                  }
                  code = -1;
                  break;
               case 'a':
                  all_objects = 1;
                  if (verbose) {
                     fprintf(stderr, "all objects will be included\n");
                  }
                  if (verbose) {
                     fprintf(stderr, "option -a implies -b\n");
                  }
                  long_array = 0;
                  break;

               case 'b':
                  long_array = 0;
                  if (verbose) {
                     fprintf(stderr, "array will be generated as bytes\n");
                  }
                  break;

               case 'c':
                  gen_function = 1;
                  tiny_only = 0;
                  if (verbose) {
                     fprintf(stderr, "C function will be generated (TINY, SMALL or LARGE mode)\n");
                  }
                  if (verbose) {
                     fprintf(stderr, "option -c implies -a and -b\n");
                  }
                  all_objects = 1;
                  long_array = 0;
                  break;

               case 'd':
                  diagnose++;   /* increase diagnosis level */
                  verbose = 1;   /* diagnose implies verbose */
                  if (diagnose == 1) {
                     fprintf(stderr, "Catalina Payload %s\n", VERSION); 
                  }
                  fprintf(stderr, "diagnostic level %d\n", diagnose);
                  break;

               case 'f':
                  func_named = 1;
                  if (strlen(argv[i]) == 2) {
                     if (argc > 0) {
                        // use next arg
                        safecpy(func_name,"", MAX_LINELEN);
                        pathcat(func_name, argv[++i], NULL, MAX_LINELEN);
                        argc--;
                     }
                     else {
                        fprintf(stderr, "option -f requires an argument\n");
                        code = -1;
                        break;
                     }
                  }
                  else {
                     // use remainder of this arg
                     safecpy(func_name,"", MAX_LINELEN);
                     pathcat(func_name, &argv[i][2], NULL, MAX_LINELEN);
                  }
                  if (verbose) {
                     fprintf(stderr, "function name = %s\n", func_name);
                  }
                  break;

               case 'l':
                  long_array = 1;
                  if (verbose) {
                     fprintf(stderr, "array will be generated as longs\n");
                  }
                  break;

               case 'n':
                  named = 1;
                  if (strlen(argv[i]) == 2) {
                     if (argc > 0) {
                        // use next arg
                        safecpy(array_name,"", MAX_LINELEN);
                        pathcat(array_name, argv[++i], NULL, MAX_LINELEN);
                        argc--;
                     }
                     else {
                        fprintf(stderr, "option -n requires an argument\n");
                        code = -1;
                        break;
                     }
                  }
                  else {
                     // use remainder of this arg
                     safecpy(array_name,"", MAX_LINELEN);
                     pathcat(array_name, &argv[i][2], NULL, MAX_LINELEN);
                  }
                  if (verbose) {
                     fprintf(stderr, "array name = %s\n", array_name);
                  }
                  break;

               case 's':
                  if (strlen(argv[i]) == 2) {
                     // use next arg
                     if (argc > 0) {
                        sscanf(argv[++i], "%d", &stack_size);
                        argc--;
                     }
                     else {
                        fprintf(stderr, "option -s requires an argument\n");
                        code = -1;
                        break;
                     }
                  }
                  else {
                     // use remainder of this arg
                     sscanf(&argv[i][2], "%d", &stack_size);
                  }
                  if (verbose) {
                     fprintf(stderr, "stack size = %d\n", stack_size);
                  }
                  break;

               case 't':
                  gen_function = 1;
                  tiny_only = 1;
                  if (verbose) {
                     fprintf(stderr, "C function will be generated (TINY mode only)\n");
                  }
                  if (verbose) {
                     fprintf(stderr, "option -t implies -a and -b\n");
                  }
                  all_objects = 1;
                  long_array = 0;
                  break;

               case 'v':
                  verbose = 1;
                  fprintf(stderr, "verbose mode\n");
                  break;

               default:
                  fprintf(stderr, "\nError: unrecognized switch: %s\n", argv[i]);
                  code = -1; // force exit without further processing
                  break;
            }
         }
         else {
            // assume its a filename
            if (file_count < MAX_FILES) {
               file_name[file_count++] = strdup(argv[i]);
               if (diagnose) {
                  fprintf(stderr, "file %s\n", argv[i]);
               }
               code = 1; // work to do
            }
            else {
               fprintf(stderr, "\ntoo many files specified - file %s will be ignored\n", argv[i]);
            }
         }
      }
      i++; // next argument
   }
   if (code == -1) {
      return code;
   }
   if (diagnose) {
      fprintf(stderr, "executable name = %s\n", argv[0]);
   }
   if ((argc == 1) || (file_count == 0)) {
      if (strlen(argv[0]) == 0) {
         // in case my name was not passed in ...
         help("spinc");
      }
      else {
         help(argv[0]);
      }
      code = -1;
   }

   return code;

}


/*
 * load spin binary from filename
 * @param name - path and filename
 * @returns pointer to binary or 0 on failure.
 */
static char* load_binary(char* name)
{
    FILE* fp;
    char* rp = 0;
    int   size = 0;
    int   ii;
    char* bin = 0;
    char  hdr[sizeof(SpinHdr_st)];
    SpinHdr_st* sp = 0;

    fp = fopen(name, "rb");
    if(!fp) {
        fprintf(stderr, "Can't find spin binary '%s'.\n", name);
        goto exit;
    }
    else {
        for(ii = 0; ii < 16; ii++) {
            hdr[ii] = fgetc(fp);
            if (feof(fp)) {
                fprintf(stderr, "Bad spin binary header. Unexpected end of file in '%s'\n", name);
                goto exit;
            }
        }
        sp = (SpinHdr_st*) hdr;
        if(sp->objstart != SPIN_OBJSTART) {
            fprintf(stderr, "Bad spin object start in '%s'. Expecting a spin binary for conversion.\n", name);
            goto exit;
        }
        size = sp->varstart;
        fclose(fp);
        fp = fopen(name, "rb");
        rp = (char*) malloc(size);
//        for(ii = 0; ii < size; ii++) {
//            rp[ii] = fgetc(fp);
//        }
        ii = fread(rp,1,size,fp);
        if (ii <= 0) {
            fprintf(stderr, "Can't read spin binary '%s' [%d].\n", name, ii);
            goto exit;
        }
    }
exit:
    if(fp) fclose(fp);
    return rp;
}


/*
 * print array of C longs
 * @param p - pointer to spin binary image
 * @param name - array name
 */
static void print_long_array(char* p, char* name)
{
    int ii;
    int datstart = 0;
    int datlen = 0;
    SpinHdr_st* sp = (SpinHdr_st*) p;
    SpinObj_st* op = (SpinObj_st*) &sp->code;
    int hdrlen = ((int)op-(int)sp);

    if (all_objects) {
       datstart = hdrlen;
       datlen   = sp->varstart-hdrlen;
    }
    else {
       datstart = hdrlen+(op->objcnt+op->pubcnt)*4;
       datlen   = sp->pubstart-datstart;
    }
    printf("unsigned long %s_array[] =\n{\n    ", name);
    for (ii = 0; ii < datlen; ii += 4) {
        if (ii && ((ii % 16) == 0)) {
           if (diagnose) {
              printf("// 0x%04x",ii+datstart-16);
           }
           printf("\n    ");
        }
        printf("0x%08x", *(int*)(p+ii+datstart));
        if (ii < datlen - 4) {
           printf(", ");
        }
    }
    printf("\n};\n");
}


/*
 * print array of C bytes (chars)
 * @param p - pointer to spin binary image
 * @param name - array name
 */
static void print_byte_array(char* p, char* name)
{
    int ii;
    int datstart = 0;
    int datlen = 0;
    SpinHdr_st* sp = (SpinHdr_st*) p;
    SpinObj_st* op = (SpinObj_st*) &sp->code;
    int hdrlen = ((int)op-(int)sp);

    if (all_objects) {
       datstart = hdrlen;
       datlen   = sp->varstart-hdrlen;
    }
    else {
       datstart = hdrlen+(op->objcnt+op->pubcnt)*4;
       datlen   = sp->pubstart-datstart;
    }
    printf("unsigned char %s_array[] =\n{\n    ", name);
    for (ii = 0; ii < datlen; ii++) {
        if (ii && ((ii % 8) == 0)) {
           if (diagnose) {
              printf("// 0x%04x",ii+datstart-8);
           }
           printf("\n    ");
        }
        printf("0x%02x", *(unsigned char*)(p+ii+datstart));
        if (ii < datlen - 1) {
           printf(", ");
        }
    }
    printf("\n};\n");
}


/*
 * print SPIN information as #defines
 * @param p - pointer to spin binary image
 * @param name - array name
 */
static void print_spin_defines(char* p, char* name)
{
    int datstart = 0;
    int datlen = 0;
    SpinHdr_st* sp = (SpinHdr_st*) p;
    SpinObj_st* op = (SpinObj_st*) &sp->code;
    int hdrlen = ((int)op-(int)sp);

    datstart = (int) (hdrlen+(op->objcnt+op->pubcnt)*4);

    printf("\n");
    printf("#define %s_CLOCKFREQ  %d\n", name, sp->clkfreq);
    printf("#define %s_CLOCKMODE  0x%02X\n", name, sp->clkmode);
    printf("\n");

    if (all_objects) {
       printf("#define %s_ARRAY_TYPE 1 // array contains SPIN program\n", name);
       printf("#define %s_PROG_SIZE  %d // bytes\n", name, sp->varstart-hdrlen);
       printf("#define %s_PROG_OFF   %d // byte offset in array\n", name, sp->pubstart-hdrlen);
       printf("#define %s_VAR_SIZE   %d // bytes\n", name, sp->stkstart-sp->varstart-8);
       printf("#define %s_STACK_SIZE %d // bytes\n", name, stack_size);
       printf("#define %s_STACK_OFF  %d // bytes\n", name, sp->stkbase-sp->stkstart);
    }
    else {
       printf("#define %s_ARRAYTYPE 0 // array contains PASM program\n", name);
       printf("#define %s_PROG_SIZE  %d // bytes\n", name, sp->pubstart-datstart);
    }
    printf("\n");
}


/*
 * print function to call
 * @param name - array name
 */
static void print_function(char* name)
{
   if (tiny_only) {
      // do not need a prog array - only var and stack
      printf("\n");
      printf("int start_%s(void *var, void *stack) {\n\n", name);
      printf("   // zero the memory used as var and stack space\n");
      printf("   memset(var, 0, %s_VAR_SIZE);\n", name);
      printf("   memset(stack, 0, %s_STACK_SIZE);\n", name);
      printf("\n");
      printf("   // start the Spin program\n");
      if (func_named) {
         printf("   return %s(%s_array, var, stack, %s_PROG_OFF, %s_STACK_OFF);\n", func_name, name, name, name);
      }
      else {
         printf("   return _coginit_Spin(%s_array, var, stack, %s_PROG_OFF, %s_STACK_OFF);\n", name, name, name);
      }
      printf("\n");
      printf("}\n\n");
   }
   else {
      // need a prog array that is in Hub RAM
      printf("\n");
      printf("int start_%s(void *prog, void *var, void *stack) {\n\n", name);
      printf("   // zero the memory used as var and stack space\n");
      printf("   memset(var, 0, %s_VAR_SIZE);\n", name);
      printf("   memset(stack, 0, %s_STACK_SIZE);\n", name);
      printf("\n");
      printf("   // copy the Spin program to the local array, in case it is\n");
      printf("   // currently stored in XMM RAM\n");
      printf("   memcpy(prog, %s_array, %s_PROG_SIZE);\n", name, name);
      printf("\n");
      printf("   // start the Spin program\n");
      if (func_named) {
         printf("   return %s(%s_array, var, stack, %s_PROG_OFF, %s_STACK_OFF);\n", func_name, name, name, name);
      }
      else {
         printf("   return _coginit_Spin(%s_array, var, stack, %s_PROG_OFF, %s_STACK_OFF);\n", name, name, name);
      }
      printf("\n");
      printf("}\n\n");
   }
}


/*
 * chop path from filename
 * @param name - path and filename
 * @returns filename
 */
static char* chopPath(char* name)
{
    int ii = 0;
    int len = strlen(name);
    char* start = name;
    char* p = 0;
    int sep = 0;

    if(strchr(name, '\\'))  sep = '\\';
    if(strchr(name, '/'))   sep = '/';

    while((start+len-name) > 0) {
        p = strchr(name+1, sep);
        if (!p) break;
        name = p+1;
    }
    return name;
}


/*
 * chop extension from filename
 * @param name - path and filename
 * @returns filename
 */
static char* chopExtension(char* name)
{
    char* p  = strchr(name, '.');
    int   ii = (int)p-(int)name;
    if (p && ii) name[ii] = '\0';
    return name;
}

int main(int argc, char* argv[])
{
    char * spin_ptr = 0;
    char spin_name[MAX_LINELEN + 1] = "";
    char spin_num[5] = "";
    int    i;

   if (decode_arguments(argc, argv) <= 0) {
      if (diagnose) {
         fprintf(stderr, "%s exiting\n", argv[0]);
      }
      exit(0);
   }

   if (file_count == 0) {
      fprintf(stderr, "\nNo input files specified - exiting\n");
      exit(1);
   }

   for (i = 0; i < file_count; i++) {

      spin_ptr = load_binary(file_name[i]);
      if (spin_ptr) {
          printf("/* \n");
          printf(" * Created from %s", file_name[i]);
          if (gen_function) {
             printf (" with option -c\n"); 
          }
          else if (all_objects) {
             printf (" with option -a\n"); 
          }
          else {
             printf("\n");
          }
          printf(" * \n");
          printf(" * by Spin to C Array Converter, version %s\n", VERSION);
          printf(" */\n");
          if (named) {
             safecpy(spin_name, array_name, MAX_LINELEN);
             if (file_count > 1) {
                sprintf(spin_num, "_%d", i);
                safecat(spin_name, spin_num, MAX_LINELEN);
             }
          }
          else {
             safecpy(spin_name, chopExtension(file_name[i]), MAX_LINELEN);
          }
          print_spin_defines(spin_ptr, spin_name);
          if (long_array) {
             print_long_array(spin_ptr, spin_name);
          }
          else {
             print_byte_array(spin_ptr, spin_name);
          }
          if (gen_function) {
             print_function(spin_name);
          }
      }
   }
   return 0;
}

